/* 
 *   DTMF decode module, part of:
 *      BSD Telephony Of Mexico "Zapata" Telecom Library, version 1.5  1/1/01
 *
 *   Part of the "Zapata" Computer Telephony Technology.
 *
 *   See http://www.bsdtelephony.com.mx
 *
 *
 *  The technologies, software, hardware, designs, drawings, scheumatics, board
 *  layouts and/or artwork, concepts, methodologies (including the use of all
 *  of these, and that which is derived from the use of all of these), all other
 *  intellectual properties contained herein, and all intellectual property
 *  rights have been and shall continue to be expressly for the benefit of all
 *  mankind, and are perpetually placed in the public domain, and may be used,
 *  copied, and/or modified by anyone, in any manner, for any legal purpose,
 *  without restriction.
 *
 *   This module heavily modified, but based upon works of authors referenced
 *   below.
 */


/* Heavily modified by Mark Balliet to include it as a realtime
   DTMF detection routine for the Linux VM100 project        */

/* a proof of concept for DTMF detection with the goertzel algo
   written 1996 by christian mock <cm@kukuruz.ping.at>
   placed in the public domain.
   siehe:
   http://ptolemy.eecs.berkeley.edu/~pino/Ptolemy/papers/96/dtmf_ict/
   */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>
#include <stdio.h>
#include <tor.h>

#define	PI 3.1414926535

#include "goertzel.h"
#include "zap.h"

int mulaw[256];
int fullmulaw[256];
int cos2pik[NCOEFF];      /* 2 * cos(2 * PI * k / N), precalculated */
int mcos2pik[MNCOEFF];  /* same for MF */
 
/*--------------------------------------------------------------------*/
void goertzel_init(void)
{
   int i, kp, k;
   /* 
    *  Set up conversion table
    */
   for(i=0;i<256;i++)
   {
      short mu,e,f,y;
      short etab[]={0,132,396,924,1980,4092,8316,16764};

      mu=255-i;
      e=(mu&0x70)/16;
      f=mu&0x0f;
      y=f*(1<<(e+3));

      y=etab[e]+y;
      if(mu&0x80)
	 y=-y;

        fullmulaw[i] = y;
        mulaw[i]= (y >> (15 - AMP_BITS));
   }
   /* precalculate 2 * cos (2 PI k / N) */
   kp = 0;
   for(i = 0; i < 8; i++) {
     k = (int)rint((double)dtmf_tones[i].freq * NPOINTS / SAMPLERATE);
     cos2pik[kp] = 2 * (32768 * cos(2 * PI * k / NPOINTS));
     dtmf_tones[i].k = kp++;
     
     k = (int)rint((double)dtmf_tones[i].freq * 2 * NPOINTS / SAMPLERATE);
     cos2pik[kp] = 2 * (32768 * cos(2 * PI * k / NPOINTS));
     dtmf_tones[i].k2 = kp++;
   }
   kp = 0;
   for(i = 0; i < 6; i++) {
     k = (int)rint((double)mf_tones[i].freq * NPOINTS / SAMPLERATE);
     mcos2pik[kp] = 2 * (32768 * cos(2 * PI * k / NPOINTS));
     mf_tones[i].k = kp++;
     
     k = (int)rint((double)mf_tones[i].freq * 2 * NPOINTS / SAMPLERATE);
     mcos2pik[kp] = 2 * (32768 * cos(2 * PI * k / NPOINTS));
     mf_tones[i].k2 = kp++;
   }
}

/*--------------------------------------------------------------------*/
int get_buffer(int fd, char *readbuffer, int buffer[], int samples) {
  int nread, i,j;
  int *ptr;
  unsigned char *ptr2;
  
  nread = read(fd, readbuffer, samples);
  if (nread < samples) return(-1);

  ptr = buffer;
  ptr2 = readbuffer;
  
  for(i = 0; i < samples; i++)
    *(ptr++) = (mulaw[*(ptr2++)] / 3) * 2;

  return nread;
}


/*--------------------------------------------------------------------*/
int goertzel(int N, int lohi, int *sample, int *cos2pik, int *result) {
  int sk, sk1, sk2;
  int k, n;

 if (lohi < 2)
 {
  for(k = 8 * lohi; k < 8 * (lohi + 1); k++) {
    sk = sk1 = sk2 = 0.0;
    for(n = 0; n < N; n++) {
      sk = sample[n] + (((int)(cos2pik[k] * sk1)) >> 15) - sk2;
      sk2 = sk1;
      sk1 = sk;
    }
    result[k] =
      ((sk * sk) >> AMP_BITS) - (((((int)(cos2pik[k] * sk)) >> 15) * sk2) >> AMP_BITS) +
      ((sk2 * sk2) >> AMP_BITS);
/*    if (result[k] < 0) result[k] = -result[k]; */
  }
 }
 else
 {
  for(k = 0; k < MNCOEFF; k++) {
    sk = sk1 = sk2 = 0.0;
    for(n = 0; n < N; n++) {
      sk = sample[n] + ((cos2pik[k] * sk1) >> 15) - sk2;
      sk2 = sk1;
      sk1 = sk;
    }
    result[k] =
      ((sk * sk) >> AMP_BITS) - ((((cos2pik[k] * sk) >> 15) * sk2) >> AMP_BITS) +
      ((sk2 * sk2) >> AMP_BITS);
   if (result[k] < 0) result[k] = -result[k];
  }
 }
 return 0;
}

/* Low DTMF group filter.. 5th order Butterworth Bandpass filter with corners
650 and 1075 hz, with extra zero at 1209 hz. Hi group no less then 40db down.

/* Digital filter designed by mkfilter/mkshape/gencode   A.J. Fisher
   Command line: /www/usr/fisher/helpers/mkfilter -Bu -Bp -o 5 -a 8.1250000000e-02 1.3437500000e-01 -Z 1.5112500000e-01 -l */

#define LNZEROS 12
#define LNPOLES 12
#define LGAIN   5.002691440e+03

int lowfilter(ZAP *zap,int val)
      { 
	zap->lxv[0] = zap->lxv[1]; zap->lxv[1] = zap->lxv[2]; zap->lxv[2] = zap->lxv[3]; zap->lxv[3] = zap->lxv[4]; zap->lxv[4] = zap->lxv[5]; zap->lxv[5] = zap->lxv[6]; zap->lxv[6] = zap->lxv[7]; zap->lxv[7] = zap->lxv[8]; zap->lxv[8] = zap->lxv[9]; zap->lxv[9] = zap->lxv[10]; zap->lxv[10] = zap->lxv[11]; zap->lxv[11] = zap->lxv[12]; 
        zap->lxv[12] = ((float)val) / LGAIN;
        zap->lyv[0] = zap->lyv[1]; zap->lyv[1] = zap->lyv[2]; zap->lyv[2] = zap->lyv[3]; zap->lyv[3] = zap->lyv[4]; zap->lyv[4] = zap->lyv[5]; zap->lyv[5] = zap->lyv[6]; zap->lyv[6] = zap->lyv[7]; zap->lyv[7] = zap->lyv[8]; zap->lyv[8] = zap->lyv[9]; zap->lyv[9] = zap->lyv[10]; zap->lyv[10] = zap->lyv[11]; zap->lyv[11] = zap->lyv[12]; 
        zap->lyv[12] =   (zap->lxv[12] - zap->lxv[0]) +   1.1641040230 * (zap->lxv[1] - zap->lxv[11]) + 4 * (zap->lxv[2] - zap->lxv[10])
                     +   5.8205201151 * (zap->lxv[9] - zap->lxv[3]) + 5 * (zap->lxv[8] - zap->lxv[4]) +  11.6410402300 * (zap->lxv[5] - zap->lxv[7])
                     + ( -0.3374167881 * zap->lyv[2]) + (  2.9538237865 * zap->lyv[3])
                     + (-12.4237370100 * zap->lyv[4]) + ( 32.7069548080 * zap->lyv[5])
                     + (-59.4151801510 * zap->lyv[6]) + ( 77.6350642960 * zap->lyv[7])
                     + (-73.8487784780 * zap->lyv[8]) + ( 50.5296515660 * zap->lyv[9])
                     + (-23.8573288510 * zap->lyv[10]) + (  7.0493879100 * zap->lyv[11]);
        return((int)zap->lyv[12]);
      }

/* High DTMF group filter.. 3rd order Butterworth Bandpass filter with corners
1125hz and 1580 hz, with extra zero at 941 hz. Lo group no less then 60db down.

/* Digital filter designed by mkfilter/mkshape/gencode   A.J. Fisher
   Command line: /www/usr/fisher/helpers/mkfilter -Bu -Bp -o 3 -a 1.4062500000e-01 1.9750000000e-01 -Z 1.1762500000e-01 -l */

#define HNZEROS 8
#define HNPOLES 8
#define HGAIN   1.226388263e+02

int highfilter(ZAP *zap,int val)
      { 
	zap->hxv[0] = zap->hxv[1]; zap->hxv[1] = zap->hxv[2]; zap->hxv[2] = zap->hxv[3]; zap->hxv[3] = zap->hxv[4]; zap->hxv[4] = zap->hxv[5]; zap->hxv[5] = zap->hxv[6]; zap->hxv[6] = zap->hxv[7]; zap->hxv[7] = zap->hxv[8]; 
        zap->hxv[8] = ((float) val) / HGAIN;
        zap->hyv[0] = zap->hyv[1]; zap->hyv[1] = zap->hyv[2]; zap->hyv[2] = zap->hyv[3]; zap->hyv[3] = zap->hyv[4]; zap->hyv[4] = zap->hyv[5]; zap->hyv[5] = zap->hyv[6]; zap->hyv[6] = zap->hyv[7]; zap->hyv[7] = zap->hyv[8]; 
        zap->hyv[8] =   (zap->hxv[8] - zap->hxv[0]) +   1.4782045682 * (zap->hxv[1] - zap->hxv[7]) + 2 * (zap->hxv[2] - zap->hxv[6])
                     +   4.4346137047 * (zap->hxv[5] - zap->hxv[3])
                     + ( -0.4874114522 * zap->hyv[2]) + (  1.6186021854 * zap->hyv[3])
                     + ( -3.6131601704 * zap->hyv[4]) + (  4.7316679780 * zap->hyv[5])
                     + ( -4.5867417716 * zap->hyv[6]) + (  2.6168057691 * zap->hyv[7]);
        return((int)zap->hyv[8]);
      }


/*--------------------------------------------------------------------*/
int goertzel_link(ZAP *zap,int inf, char *readbuffer, int mode, char *last,int *nsamp, int tedge)
{
  int silence, grp[2], flr[2];
  
  int buffer[NPOINTS],buffer1[NPOINTS];
  int result[NCOEFF];

  int i,j,retval,m,maybe,confn,confm;
  char what;

 retval = 0;
 maybe = 0;
for(;;)
 {
  i = get_buffer(inf, readbuffer,buffer, NPOINTS);
  if (i < 0) return(i);
 
  for(i = 0,j = 0; i < NPOINTS; i++)
    {
	if ((buffer[i] < SILENCE_LEVEL) &&
		(buffer[i] > -SILENCE_LEVEL)) j++; else j = 0;
    }

  if (j) zap->nsilence += j; else zap->nsilence = 0;

    /* if counting, add sample count */
  if (nsamp) *nsamp += NPOINTS; 

  if (!mode) /* if DTMF */
    {
	  for(i = 0; i < NPOINTS; i++) buffer1[i] = lowfilter(zap,buffer[i]);

	  goertzel(NPOINTS, 0, buffer1, cos2pik, result);


	  for(i = 0; i < NPOINTS; i++) buffer1[i] = highfilter(zap,buffer[i]);

	  goertzel(NPOINTS, 1, buffer1, cos2pik, result);

	#ifdef	DISPLAYRESULT
	  for(i = 0; i < 8; i++)
		printf("result [%d] is %d %d\n",i,result[i * 2],result[1 + (i * 2)]);
	#endif

	  grp[0] = grp[1] = -2;
	  silence = 0;

	  for(flr[0] = 0,flr[1] = 0,i = 0; i < 8; i++)
	   {
		flr[dtmf_tones[i].grp] += result[dtmf_tones[i].k];
	   }
	   flr[0] = (flr[0] * 93) / 100;  /* must be at least 93% of total */
	   flr[1] = (flr[1] * 93) / 100;  /* must be at least 93% of total */

	  if (flr[0] < DTMF_TRESH) flr[0] = DTMF_TRESH;
	  if (flr[1] < DTMF_TRESH) flr[1] = DTMF_TRESH;

	  for(i = 0; i < 8; i++)
	    if((result[dtmf_tones[i].k] > flr[dtmf_tones[i].grp]) &&
	       (result[dtmf_tones[i].k2] < H2_TRESH))
	    	grp[dtmf_tones[i].grp] = (grp[dtmf_tones[i].grp] == -2  ? i : -1);
	    else
	    if((result[dtmf_tones[i].k] < SILENCE_TRESH) &&
	       (result[dtmf_tones[i].k2] < SILENCE_TRESH))
	        silence++;
	    if(silence == 8)
	      what = ' ';
	    else {
	      if((grp[LOGRP] >= 0) && (grp[HIGRP] >= 0)) {
		what = dtmf_matrix[grp[LOGRP]][grp[HIGRP] - 4];
		if(*last != ' ' && *last != '.')
		  *last = what;		/* min. 1 period of non-DTMF between tones */
	      } else
		what = '.';
	    }
	    if((what != *last)&&(what !='.')&&(what !=' '))
		{
	#ifdef	DISPLAYOUT
			printf("%c",what);fflush(stdout);
	#endif
			retval = what;
		}
     }
    else /* if MF */
     {

	  goertzel(NPOINTS, 2, buffer, mcos2pik, result);

	#ifdef	DISPLAYRESULT
	  for(i = 0; i < 6; i++)
		printf("result [%d] is %d %d\n",i,result[i * 2],result[1 + (i * 2)]);
	#endif

	    /* get highest */
	  for(flr[0] = 0,i = 0; i < 6; i++)
	   {
		if (result[mf_tones[i].k] > flr[0])
			 flr[0] = result[mf_tones[i].k];
	   }
	flr[0] = (flr[0] * 45) / 100;
	if (zap->mflevel) 
	   {
		flr[0] = (flr[0] + zap->mflevel) / 2;
	   }
	if (flr[0] < MF_TRESH) flr[0] = MF_TRESH;
	zap->mflevel = flr[0];
	flr[1] = (flr[0] * 30) / 100;

	  m = 0;
	  silence = 0;
	  for(i = 0; i < 6; i++)
	   {
	    if ((result[mf_tones[i].k] > /* MF_TRESH */ flr[0]) && 
		(result[mf_tones[i].k2] < /* MF_H2_TRESH */ flr[1])) 
		   {
			m |= 1 << i;
		   }
	    else 
	    if((result[mf_tones[i].k] < SILENCE_TRESH) &&
	       (result[mf_tones[i].k2] < SILENCE_TRESH))
	        silence++;
	  }
	 if (silence == 6) what = ' ';  /* all is silent */
	    else { /* we have something */
	      if (mfs[m] != ' ') { /* if what we have is valid */
		what = mfs[m]; /* get from decode string */
		if(*last != ' ' && *last != '.')
		  *last = what;		/* min. 1 period of non-DTMF between tones */
	      } else
		what = '.';
	    }
	    if((what != *last)&&(what !='.')&&(what !=' '))
		{
	#ifdef	DISPLAYOUT
			printf("%c",what);fflush(stdout);
	#endif
			retval = what;
		}
     }
    *last = what;
    if (retval && (retval == what)) /* if tone decoded */
      {
	if ((!maybe) && (zap->digit_mode & ZAP_MUTECONF)) 
	   {
		i = 1; /* mute off conf */
		if (ioctl(zap->fd,TOR_CONFMUTE,&i) == -1) return(-1);
		maybe = 1;
		*last = ' ';
		retval = 0;
		continue;
	   }
	if (tedge) continue; /* wait until end of toneage */
      }
    i = 0;  /* put back on conf */
    if (ioctl(zap->fd,TOR_CONFMUTE,&i) == -1) return(-1);
    return(retval);
  }
}
/*--------------------------------------------------------------------*/
/*--------------------------------------------------------------------*/

